#[non_exhaustive]pub struct BlockStreamStats {
pub block_count: u64,
pub block_shift: u32,
pub data_capacity: usize,
pub data_used: usize,
pub data_wasted: usize,
pub data_available: usize,
}Expand description
Stats about an individual block::Stream.
Fields (Non-exhaustive)
This struct is marked as non-exhaustive
Non-exhaustive structs could have additional fields added in future. Therefore, non-exhaustive structs cannot be constructed in external crates using the traditional
Struct { .. } syntax; cannot be matched against without a wildcard ..; and struct update syntax will not work.block_count: u64The number of blocks backing up a block stream.
block_shift: u32The size of a single block expressed as a power of two.
data_capacity: usizeThe capacity for data in bytes.
data_used: usizeThe size of appended and synced data in bytes.
data_wasted: usizeThe size of inaccessible data in bytes.
data_available: usizeThe size in bytes available for appends.
Implementations
Sourceimpl BlockStreamStats
impl BlockStreamStats
Sourcepub fn blocks_size(&self) -> u64
pub fn blocks_size(&self) -> u64
Returns the size in bytes of the underlying Blocks.